from cubicweb.devtools.testlib import AutomaticWebTest


class AutomaticWebTest(AutomaticWebTest):
    def to_test_etypes(self):
        return set()

    def list_startup_views(self):
        return ("registration",)


if __name__ == "__main__":
    import unittest

    unittest.main()



# pylint: disable-msg=W0622
"""cubicweb-registration application packaging information"""

modname = "registration"
distname = "cubicweb-registration"

numversion = (0, 10, 0)
version = ".".join(str(num) for num in numversion)

license = "LGPL"
description = "public registration component for the CubicWeb framework"
author = "LOGILAB S.A. (Paris, FRANCE)"
author_email = "contact@logilab.fr"
web = "https://forge.extranet.logilab.fr/cubicweb/cubes/%s" % distname
classifiers = [
    "Environment :: Web Environment",
    "Framework :: CubicWeb",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: JavaScript",
]

__depends__ = {
    "cubicweb": ">= 3.27.0, < 3.34.0",
    "cubicweb-forgotpwd": None,
    "Pillow": None,
}



# copyright 2009-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""anonymous registration form and views"""

from yams.schema import role_name

from cubicweb import mail, crypto
from cubicweb.view import StartupView
from cubicweb.predicates import anonymous_user
from cubicweb.web import Redirect, ValidationError, ProcessFormError
from cubicweb.web import controller, form, captcha
from cubicweb.web import formwidgets as fw, formfields as ff
from cubicweb.web.views import forms, basecomponents, urlrewrite

from cubicweb import _


def qname(attr):
    return role_name(attr, "subject")


class RegistrationForm(forms.FieldsForm):
    __regid__ = "registration"
    domid = "registrationForm"
    form_buttons = [fw.SubmitButton()]

    @property
    def action(self):
        return self._cw.build_url("registration_sendmail")

    # properly name fields according to validation errors that may be raised by
    # the register_user service
    login = ff.StringField(
        widget=fw.TextInput(),
        role="subject",
        # we don't want to see 'authenticate'
        label=_("i18n_register_login"),
        required=True,
    )
    upassword = ff.StringField(widget=fw.PasswordInput(), role="subject", required=True)
    email_address = ff.StringField(
        widget=fw.TextInput(),
        role="subject",
        required=True,
        label=_("i18n_email_address"),
    )
    firstname = ff.StringField(widget=fw.TextInput(), role="subject")
    surname = ff.StringField(widget=fw.TextInput(), role="subject")
    captcha = ff.StringField(
        widget=captcha.CaptchaWidget(),
        required=True,
        label=_("captcha"),
        help=_("please copy the letters from the image"),
    )


class RegistrationFormView(form.FormViewMixIn, StartupView):
    __regid__ = "registration"

    def call(self):
        form = self._cw.vreg["forms"].select("registration", self._cw)
        form.render(w=self.w, display_progress_div=False)


class RegistrationSendMailController(controller.Controller):
    __regid__ = "registration_sendmail"
    content = _(
        """
Hello %(firstname-subject)s %(surname-subject)s,

thanks for registering on %(base_url)s.

Please click on the link below to activate your account :
%(url)s

See you soon on %(base_url)s !
"""
    )
    subject = _("Confirm your registration on %(base_url)s")

    def publish(self, rset=None):
        data = self.checked_data()
        recipient = data[qname("email_address")]
        msg = self.build_email(recipient, data)
        self._cw.vreg.config.sendmails([(msg, (recipient,))])
        raise Redirect(self.success_redirect_url())

    def checked_data(self):
        """only basic data check here (required attributes and password
        confirmation check)
        """
        form = self._cw.vreg["forms"].select("registration", self._cw)
        form.formvalues = {}  # init fields value cache
        data = {}
        errors = {}
        for field in form.fields:
            try:
                for field, value in field.process_posted(form):
                    if value is not None:
                        data[field.role_name()] = value
            except ProcessFormError as exc:
                errors[field.role_name()] = str(exc)
        if errors:
            raise ValidationError(None, errors)
        return data

    def build_email(self, recipient, data):
        activationurl = self.activation_url(data)  # build url before modifying data
        data.setdefault(qname("firstname"), "")
        data.setdefault(qname("surname"), "")
        if not (data.get(qname("firstname")) or data.get(qname("surname"))):
            data[qname("firstname")] = data[qname("login")]
        data.update({"base_url": self._cw.base_url(), "url": activationurl})
        content = self._cw._(self.content) % data
        subject = self._cw._(self.subject) % data
        return mail.format_mail(
            {},
            [recipient],
            content=content,
            subject=subject,
            config=self._cw.vreg.config,
        )

    def activation_url(self, data):
        data.pop(qname("upassword") + "-confirm", None)
        key = crypto.encrypt(data, self._cw.vreg.config["registration-cypher-seed"])
        return self._cw.build_url("registration_confirm", key=key, __secure__=True)

    def success_redirect_url(self):
        msg = self._cw._(
            "Your registration email has been sent. Follow "
            "instructions in there to activate your account."
        )
        return self._cw.build_url("", __message=msg)


class RegistrationConfirmController(controller.Controller):
    __regid__ = "registration_confirm"

    def publish(self, rset=None):
        req = self._cw
        try:
            data = crypto.decrypt(
                req.form["key"], req.vreg.config["registration-cypher-seed"]
            )
            login = data[qname("login")]
            password = data.pop(qname("upassword"))
        except Exception:
            msg = req._("Invalid registration data. Please try registering again.")
            raise Redirect(req.build_url("register", __message=msg))
        if self._cw.user.login == login:
            # already logged in (e.g. regstration link replayed twice in the browser)
            raise Redirect(self.success_redirect_url(self._cw.user.name()))
        req.form = data  # hijack for proper validation error handling
        err = None
        try:
            with self.appli.repo.internal_cnx() as cnx:
                cnx.call_service(
                    "register_user",
                    login=login,
                    password=password,
                    email=data.get(qname("email_address")),
                    firstname=data.get(qname("firstname")),
                    surname=data.get(qname("surname")),
                )
                cnx.commit()
        except ValidationError as validation_error:
            err = validation_error
            if "login" in err.errors and "unicity constraint" in err.errors["login"]:
                err.errors.pop("", None)
                err.errors.pop("login", None)
                err.errors["login-subject"] = _(
                    'the value "{}" is already used, use another one'
                ).format(login)
                err.entity = None
            # XXX TEMPORARY HACK to allow registration links to work more than
            # once. This is required because some email clients (e.g. kmail)
            # start by downloading the url to find the mimetype of the resource
            # and then execute the appropriate action (e.g. open the url in the
            # default browser) based on the mimetype.
            if list(err.errors.keys()) != ["login-subject"]:
                raise
        # try to connect using the provided credentials
        try:
            from cubicweb import repoapi

            cnx = repoapi.connect(self.appli.repo, login, password=password)
            with cnx:
                name = cnx.user.name()
            raise Redirect(self.success_redirect_url(name))
        except Exception:
            if err is not None:
                # both registration and login failed, re-raise the previous
                # ValidationError
                raise err
            raise

    def success_redirect_url(self, name):
        login = self._cw.build_url("login")
        msg = self._cw._(
            "Congratulations, your registration is complete. "
            'You can now <a href="{}">login</a>.'
        ).format(login)
        return self._cw.build_url("", __message=msg)


class RegisterLink(basecomponents.HeaderComponent):
    __regid__ = "registration.registerlink"
    __select__ = basecomponents.HeaderComponent.__select__ & anonymous_user()
    context = "header-right"

    def render(self, w):
        self._cw.add_css("cubicweb.pictograms.css")
        w(
            '<a class="logout icon-user-add" title="%s" href="%s"></a>'
            % (
                self._cw._("i18n_register_user"),
                self._cw.build_url("register"),
            )
        )


# urls #########################################################################


class RegistrationSimpleReqRewriter(urlrewrite.SimpleReqRewriter):
    ignore_baseclass_rules = True
    rules = [
        ("/register", dict(vid="registration")),
    ]



import re
import urllib.parse
from contextlib import contextmanager

from cubicweb.crypto import encrypt, decrypt
from cubicweb.devtools.testlib import MAILBOX, CubicWebTC


class RegistrationTC(CubicWebTC):

    captcha_value = "captcha value"

    data = {
        "firstname-subject": "Toto",
        "surname-subject": "Toto",
        "email_address-subject": "toto@secondweb.fr",
        "login-subject": "toto",
        "upassword-subject": "toto",
        "upassword-subject-confirm": "toto",
        "captcha": captcha_value,
    }

    def setup_database(self):
        self.config.global_set_option("registration-cypher-seed", "dummy cypher key")
        super(RegistrationTC, self).setup_database()

    def _check_user_not_created(self):
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute(
                "CWUser X WHERE X login %(login)s",
                {"login": self.data["login-subject"]},
            )
            self.assertFalse(rset)

    def _check_error(
        self,
        req,
        path,
        expected_path="registration",
        expected_errors=None,
        expected_msg=None,
        expected_formvalues=None,
    ):
        path, params = self.expect_redirect_handle_request(req, path)
        self.assertEqual(path, expected_path)
        if expected_msg:
            self.assertMessageEqual(req, params, expected_msg)
        forminfo = req.session.data.get("registration")
        if forminfo is None:
            self.assertFalse(expected_errors or expected_formvalues)
        else:
            self.assertEqual(forminfo["eidmap"], {})
            self.assertEqual(forminfo["values"], expected_formvalues or {})
            error = forminfo["error"]
            self.assertEqual(error.entity, None)
            self.assertEqual(error.errors, expected_errors or {})

    def _posted_form(self, *skipkeys):
        data = self.data.copy()
        for key in skipkeys:
            data.pop(key, None)
        if "__errorurl" not in skipkeys:
            data["__errorurl"] = "registration"
        return data

    def test_registration_form(self):
        with self.admin_access.web_request() as req:
            req.form = {"firstname-subject": "Toto"}
            pageinfo = self.view("registration", req=req, rset=None)

        # check form field names
        names = pageinfo.etree.xpath(
            '//form[@id="registrationForm"]//input[@type!="hidden"]/@name'
        )
        self.assertEqual(set(names), set(self.data))

        # check form field value
        firstname = pageinfo.etree.xpath('//input[@name="firstname-subject"]/@value')
        self.assertEqual(firstname, [req.form["firstname-subject"]])

    def test_send_mail_ok(self):
        with self.new_access("anon").web_request() as req:
            req.form = self._posted_form()
            req.session.data["captcha"] = self.captcha_value
            path, params = self.expect_redirect_handle_request(
                req, "registration_sendmail"
            )
            self.assertEqual(path, "")
            self.assertMessageEqual(
                req,
                params,
                "Your registration email has been sent. "
                "Follow instructions in there to activate your account.",
            )
            # check email contains activation url...
            URL_RE = re.compile("(%s[^.]+)$" % self.config["base-url"], re.M)
            text = MAILBOX[-1].message.get_payload(decode=True)
            url = URL_RE.search(text).group(1)
            # ... and the registration key contains all data
            key = dict(urllib.parse.parse_qsl(urllib.parse.urlsplit(url)[3]))["key"]
            d = self._posted_form("upassword-subject-confirm")
            decrypred = decrypt(key, self.config["registration-cypher-seed"])
            decrypred["upassword-subject"] = decrypred["upassword-subject"].decode(
                "utf8"
            )
            self.assertDictEqual(decrypred, d)

    def test_send_mail_failure(self):
        for param, msg, val in (
            ("login-subject", "required field", None),
            ("captcha", "required field", None),
            ("captcha", "incorrect captcha value", "abc"),
        ):
            with self.admin_access.web_request() as req:
                if val is None:
                    req.form = self._posted_form(param)
                else:
                    req.form = self._posted_form()
                    req.form[param] = val
                req.session.data["captcha"] = self.captcha_value
                self._check_error(
                    req,
                    "registration_sendmail",
                    expected_formvalues=req.form,
                    expected_errors={param: msg},
                )
                self.assertEqual(req.session.data.get("captcha"), None)

    @contextmanager
    def _confirm_req(self, key=None, overriden={}):
        with self.new_access("anon").web_request() as req:
            data = self._posted_form("upassword-subject-confirm")
            data.update(overriden)
            if key is None:
                key = encrypt(data, self.config["registration-cypher-seed"])
            req.form = {"key": key}
            yield req

    def test_confirm_ok(self):
        with self._confirm_req() as req:
            path, params = self.expect_redirect_handle_request(
                req, "registration_confirm"
            )
            self.assertEqual(path, "")
            self.assertMessageEqual(
                req,
                params,
                "Congratulations, your registration is complete. "
                'You can now <a href="{}">login</a>.'.format(req.build_url("login")),
            )
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute(
                "Any U WHERE U login %(login)s, U firstname %(firstname)s, "
                "U surname %(surname)s, U use_email M, M address %(email_address)s",
                dict(
                    (k.replace("-subject", ""), v) for k, v in list(self.data.items())
                ),
            )
            self.assertTrue(rset)

    def test_confirm_failure_login_already_used(self):
        # try to recreate a 'admin' user.
        with self._confirm_req(overriden={"login-subject": "admin"}) as req:
            formvalues = self._posted_form(
                "upassword-subject", "upassword-subject-confirm"
            )
            formvalues["login-subject"] = "admin"
            # data = {'firstname-subject': 'Toto', 'surname-subject': 'Toto',
            #         'email_address-subject': 'toto@secondweb.fr',
            #         'login-subject': 'admin',
            #         'captcha': captcha_value}
            self._check_error(
                req,
                "registration_confirm",
                expected_formvalues=formvalues,
                expected_errors={
                    "login-subject": 'the value "admin" is already used, use another one'
                },
            )

    def test_confirm_failure_invalid_data(self):
        with self._confirm_req("dummykey") as req:
            self._check_error(
                req,
                "registration_confirm",
                "register",
                expected_msg="Invalid registration data. Please try registering again.",
            )
        self._check_user_not_created()

    def test_confirm_failure_email_already_used(self):
        with self.admin_access.web_request() as req:
            self.create_user(req, "test")
            req.execute(
                "INSERT EmailAddress X: U use_email X, X address %(email_address)s "
                'WHERE U login "test"',
                {"email_address": self.data["email_address-subject"]},
            )
            req.cnx.commit()
        with self._confirm_req() as req:
            req.form["__errorurl"] = "registration"
        self._check_user_not_created()


if __name__ == "__main__":
    import unittest

    unittest.main()



options = (
    (
        "registration-cypher-seed",
        {
            "type": "string",
            "default": "this is my dummy registration cypher seed",
            "help": "seed used to cypher registration data in confirmation email link",
            "group": "registration",
            "level": 2,
        },
    ),
)
